package org.jvnet.hifaces.components.templatedpanel.impl;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXResult;

import org.apache.commons.lang.StringUtils;
import org.jvnet.hifaces.components.templatedpanel.AbstractTemplate;
import org.jvnet.hifaces.components.templatedpanel.Template;
import org.jvnet.hifaces.util.SerializableDomElement;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Label;
import org.w3c.dom.Element;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;


public class CompiledDomElementTemplate extends AbstractTemplate implements Serializable {

  private static final long serialVersionUID = 1L;
  private transient DOMSource source;
  private final String templateName;
  private final String fragmentId;
  private transient Element element;
  private SerializableDomElement serializableElement;
  private transient Template compiledTemplate;
  private final String useFacetAttributeName;
  private static TransformerFactory TRANSFORMER_FACTORY = TransformerFactory.newInstance();

  protected Template getTemplate() {
    return compiledTemplate;
  }

  public String getTemplateName() {
    return templateName;
  }

  //
  public String getFragmentId() {
    return fragmentId;
  }

  public CompiledDomElementTemplate(String templateName, String fragmentId, Element element) {
    this(
        templateName,
        fragmentId,
        element,
        DefaultTemplateProvider.DEFAULT_USE_FACET_ATTRIBUTE_NAME);
  }

  public CompiledDomElementTemplate(
      String templateName,
      String fragmentId,
      Element element,
      String useFacetAttributeName) {
    this.useFacetAttributeName = useFacetAttributeName;
    this.templateName = templateName;
    this.fragmentId = fragmentId;
    this.element = element;
    this.serializableElement = new SerializableDomElement(element);
    this.source = new DOMSource(element);
    this.compiledTemplate = compileTemplate(source);
  }

  public void render(FacesContext facesContext, UIComponent component) throws IOException {
    getTemplate().render(facesContext, component);
  }

  private void readObject(java.io.ObjectInputStream in)
      throws IOException,
      ClassNotFoundException,
      SAXException {

    in.defaultReadObject();
    element = serializableElement.getElement();
    source = new DOMSource(element);
    compiledTemplate = compileTemplate(source);
  }

  public Source getSource() {
    return source;
  }

  public String correctName(String name) {
    return name.replace('.', '_').replace('-', '_').replace('\\', '_').replace(':', '_');
  }

  protected Template compileTemplate(Source source) {

    final String draftClassName = "Template"
        + templateName.hashCode()
        + (fragmentId == null ? "" : "$" + fragmentId)
        + System.currentTimeMillis();
    final String className = correctName(draftClassName);
    final ClassWriter cw = new ClassWriter(true);

    cw.visit(
        Opcodes.V1_2,
        Opcodes.ACC_PUBLIC + Opcodes.ACC_SUPER,
        className,
        null,
        "org/jvnet/hifaces/components/templatedpanel/AbstractTemplate",
        new String[]{ "org/jvnet/hifaces/components/templatedpanel/Template" }
        );

    {
      final MethodVisitor mv = cw.visitMethod(Opcodes.ACC_PUBLIC, "<init>", "()V", null, null);
      final Label l0 = new Label();
      mv.visitLabel(l0);
      mv.visitLineNumber(14, l0);
      mv.visitVarInsn(Opcodes.ALOAD, 0);
      mv.visitMethodInsn(
          Opcodes.INVOKESPECIAL,
          "org/jvnet/hifaces/components/templatedpanel/AbstractTemplate",
          "<init>",
          "()V");
      final Label l1 = new Label();
      mv.visitLabel(l1);
      mv.visitLineNumber(15, l1);
      mv.visitInsn(Opcodes.RETURN);
      final Label l2 = new Label();
      mv.visitLabel(l2);
      mv.visitLocalVariable(
          "this",
          "Lorg/jvnet/hifaces/components/templatedpanel/AbstractTemplate;",
          null,
          l0,
          l2,
          0);
      mv.visitMaxs(0, 0);
    }

    final Transformer transformer;//= getTransformer();

    synchronized (TRANSFORMER_FACTORY) {
      try {
        transformer = TRANSFORMER_FACTORY.newTransformer();
      }
      catch (TransformerConfigurationException tcex) {
        return null;
      }
    }

    {
      final MethodVisitor mv = cw.visitMethod(
          Opcodes.ACC_PUBLIC,
          "render",
          "(Ljavax/faces/context/FacesContext;Ljavax/faces/component/UIComponent;)V",
          null,
          new String[]{ "java/io/IOException" });

      final Label l0 = new Label();
      mv.visitLabel(l0);

      mv.visitVarInsn(Opcodes.ALOAD, 1);
      mv.visitMethodInsn(
          Opcodes.INVOKEVIRTUAL,
          "javax/faces/context/FacesContext",
          "getResponseWriter",
          "()Ljavax/faces/context/ResponseWriter;");
      mv.visitVarInsn(Opcodes.ASTORE, 3);

      final Label l1 = new Label();
      mv.visitLabel(l1);
      mv.visitInsn(Opcodes.ACONST_NULL);
      mv.visitVarInsn(Opcodes.ASTORE, 4);
      final Label l2 = new Label();

      final ContentHandler target = new MethodGeneratingContentHandler(mv, useFacetAttributeName);

      try {
        transformer.transform(getSource(), new SAXResult(target));
      }
      catch (TransformerException tex) {
        return null;
      }

      final Label l3 = new Label();
      mv.visitLabel(l3);
      mv.visitInsn(Opcodes.RETURN);
      final Label l4 = new Label();
      mv.visitLabel(l4);
      mv.visitLocalVariable("this", "L" + className + ";", null, l0, l4, 0);
      mv.visitLocalVariable("facesContext", "Ljavax/faces/context/FacesContext;", null, l0, l4, 1);
      mv.visitLocalVariable("component", "Ljavax/faces/component/UIComponent;", null, l0, l4, 2);
      mv.visitLocalVariable("responseWriter", "Ljavax/faces/context/ResponseWriter;", null, l1, l4, 3);
      mv.visitLocalVariable("facet", "Ljavax/faces/component/UIComponent;", null, l2, l4, 4);
      mv.visitMaxs(0, 0);
    }
    cw.visitEnd();

    final byte[] byteCode = cw.toByteArray();
    final Class clazz = loadClass(className, byteCode);

    if (clazz != null) {
      try {
        final Object object = clazz.newInstance();
        return (Template) object;
      }
      catch (Exception e) {
        e.printStackTrace();
        return null;
      }
    }
    else {
      return null;
    }
  }

  private Class loadClass(String className, byte[] b) {
    //override classDefine (as it is protected) and define the class.
    Class clazz = null;
    try {
      ClassLoader loader = getClass().getClassLoader();
      Class cls = Class.forName("java.lang.ClassLoader");
      java.lang.reflect.Method method = cls.getDeclaredMethod("defineClass", new Class[]{
          String.class,
          byte[].class,
          int.class,
          int.class });

      // protected method invocaton
      method.setAccessible(true);
      try {
        Object[] args = new Object[]{ className, b, new Integer(0), new Integer(b.length) };
        clazz = (Class) method.invoke(loader, args);
      }
      finally {
        method.setAccessible(false);
      }
    }
    catch (Exception e) {
      e.printStackTrace();
      return null;
    }
    return clazz;
  }

  private class MethodGeneratingContentHandler implements ContentHandler {

    private List labels = new ArrayList();
    protected int depth = 0;

    private String ident = "";

    public void down() {
      depth++;
      ident = ident + "  ";
    }

    public void up() {
      if (ignoredDepth != -1 && ignoredDepth == depth) {
        ignoredDepth = -1;
      }
      ident = ident.substring(0, ident.length() - 2);
      depth--;
    }

    protected int ignoredDepth = -1;

    protected boolean ignoring = false;

    public boolean isIgnoring() {
      return ignoredDepth != -1 && depth >= ignoredDepth;
    }

    public void ignore() {
      if (ignoredDepth == -1) {
        ignoredDepth = depth;
      }
    }

    private final MethodVisitor mv;
    private final String useFacetAttributeName;

    public MethodGeneratingContentHandler(MethodVisitor mv, String useFacetAttributeName) {
      this.mv = mv;
      this.useFacetAttributeName = useFacetAttributeName;
    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    public void endPrefixMapping(String prefix) throws SAXException {
    }

    public void startDocument() throws SAXException {

      //      mv.visitVarInsn(Opcodes.ALOAD, 3);
      //      mv.visitMethodInsn(
      //          Opcodes.INVOKEVIRTUAL,
      //          "javax/faces/context/ResponseWriter",
      //          "startDocument",
      //          "()V");
    }

    public void endDocument() throws SAXException {
      //      mv.visitVarInsn(Opcodes.ALOAD, 3);
      //      mv.visitMethodInsn(
      //          Opcodes.INVOKEVIRTUAL,
      //          "javax/faces/context/ResponseWriter",
      //          "endDocument",
      //          "()V");
    }

    public void startElement(String uri, String localName, String qName, Attributes atts)
        throws SAXException {
      down();
      final Label label = new Label();
      labels.add(label);

      if (!isIgnoring()) {
        final String ifFacet = atts.getValue("if-facet");
        if (!StringUtils.isEmpty(ifFacet)) {
          mv.visitVarInsn(Opcodes.ALOAD, 2);
          mv.visitLdcInsn(ifFacet);
          mv.visitMethodInsn(
              Opcodes.INVOKEVIRTUAL,
              "javax/faces/component/UIComponent",
              "getFacet",
              "(Ljava/lang/String;)Ljavax/faces/component/UIComponent;");
          mv.visitVarInsn(Opcodes.ASTORE, 4);
          mv.visitVarInsn(Opcodes.ALOAD, 4);
          mv.visitJumpInsn(Opcodes.IFNULL, label);
          mv.visitVarInsn(Opcodes.ALOAD, 4);
          mv.visitMethodInsn(
              Opcodes.INVOKEVIRTUAL,
              "javax/faces/component/UIComponent",
              "isRendered",
              "()Z");
          mv.visitJumpInsn(Opcodes.IFEQ, label);
        }
      }

      if (!isIgnoring()) {
        final String useFacet = atts.getValue(useFacetAttributeName);
        if (!StringUtils.isEmpty(useFacet)) {
          ignore();
          mv.visitVarInsn(Opcodes.ALOAD, 2);
          mv.visitLdcInsn(useFacet);
          mv.visitMethodInsn(
              Opcodes.INVOKEVIRTUAL,
              "javax/faces/component/UIComponent",
              "getFacet",
              "(Ljava/lang/String;)Ljavax/faces/component/UIComponent;");
          mv.visitVarInsn(Opcodes.ASTORE, 4);
          final Label ifNull = label;
          mv.visitVarInsn(Opcodes.ALOAD, 4);
          mv.visitJumpInsn(Opcodes.IFNULL, ifNull);
          mv.visitVarInsn(Opcodes.ALOAD, 1);
          mv.visitVarInsn(Opcodes.ALOAD, 4);
          mv.visitMethodInsn(
              Opcodes.INVOKESTATIC,
              "org/jvnet/hifaces/components/templatedpanel/util/TemplateUtils",
              "renderChild",
              "(Ljavax/faces/context/FacesContext;Ljavax/faces/component/UIComponent;)V");
          mv.visitLabel(ifNull);
        }
      }

      if (!isIgnoring()) {

        mv.visitVarInsn(Opcodes.ALOAD, 3);
        mv.visitLdcInsn(qName);
        mv.visitVarInsn(Opcodes.ALOAD, 2);
        mv.visitMethodInsn(
            Opcodes.INVOKEVIRTUAL,
            "javax/faces/context/ResponseWriter",
            "startElement",
            "(Ljava/lang/String;Ljavax/faces/component/UIComponent;)V");

        if (depth == 1) {
          mv.visitVarInsn(Opcodes.ALOAD, 3);
          mv.visitLdcInsn("id");
          mv.visitVarInsn(Opcodes.ALOAD, 2);
          mv.visitVarInsn(Opcodes.ALOAD, 1);
          mv.visitMethodInsn(
              Opcodes.INVOKEVIRTUAL,
              "javax/faces/component/UIComponent",
              "getClientId",
              "(Ljavax/faces/context/FacesContext;)Ljava/lang/String;");
          mv.visitInsn(Opcodes.ACONST_NULL);
          mv.visitMethodInsn(
              Opcodes.INVOKEVIRTUAL,
              "javax/faces/context/ResponseWriter",
              "writeAttribute",
              "(Ljava/lang/String;Ljava/lang/Object;Ljava/lang/String;)V");
        }

        for (int index = 0; index < atts.getLength(); index++) {
          final String attributeName = atts.getQName(index);
          final String attributeValue = atts.getValue(index);

          mv.visitVarInsn(Opcodes.ALOAD, 3);
          mv.visitLdcInsn(attributeName);
          mv.visitLdcInsn(attributeValue);
          mv.visitInsn(Opcodes.ACONST_NULL);
          mv.visitMethodInsn(
              Opcodes.INVOKEVIRTUAL,
              "javax/faces/context/ResponseWriter",
              "writeAttribute",
              "(Ljava/lang/String;Ljava/lang/Object;Ljava/lang/String;)V");
        }
      }
    }

    public void endElement(String uri, String localName, String qName) throws SAXException {
      if (!isIgnoring()) {
        mv.visitVarInsn(Opcodes.ALOAD, 3);
        mv.visitLdcInsn(qName);
        mv.visitMethodInsn(
            Opcodes.INVOKEVIRTUAL,
            "javax/faces/context/ResponseWriter",
            "endElement",
            "(Ljava/lang/String;)V");
      }
      final Label label = (Label) labels.remove(labels.size() - 1);
      mv.visitLabel(label);
      up();
    }

    public void characters(char[] ch, int start, int length) throws SAXException {
      if (!isIgnoring()) {
        final String str = toString(ch, start, length);
        write(toString(ch, start, length));
      }
    }

    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException {
      if (!isIgnoring()) {
        final String str = toString(ch, start, length);
        write(toString());
      }
    }

    public void processingInstruction(String target, String data) throws SAXException {
    }

    public void setDocumentLocator(Locator locator) {
    }

    public void skippedEntity(String name) throws SAXException {
    }

    private String toString(char[] ch, int start, int length) {
      final char[] chars = new char[length];
      System.arraycopy(ch, start, chars, 0, length);
      final String str = new String(chars);
      return str;
    }

    private void write(final String str) {
      mv.visitVarInsn(Opcodes.ALOAD, 3);
      mv.visitLdcInsn(str);
      mv.visitMethodInsn(
          Opcodes.INVOKEVIRTUAL,
          "javax/faces/context/ResponseWriter",
          "write",
          "(Ljava/lang/String;)V");
    }
  }
}