package org.jvnet.hifaces.components.templatedpanel.impl;

import javax.faces.context.FacesContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jvnet.hifaces.components.templatedpanel.Constants;
import org.jvnet.hifaces.components.templatedpanel.TemplateProvider;
import org.jvnet.hifaces.components.templatedpanel.TemplateProviderFactory;

public class DefaultTemplateProviderFactory implements TemplateProviderFactory {

	private TemplateProvider cachedTemplateProvider;

	private static Log logger = LogFactory
			.getLog(TemplateProviderFactory.class);

	public TemplateProvider getTemplateProvider(FacesContext context) {

		if (cachedTemplateProvider == null) {
			cachedTemplateProvider = new DefaultTemplateProvider();
		}
		return cachedTemplateProvider;
	}

	private static TemplateProviderFactory templateProviderFactory;

	public static TemplateProviderFactory getInstance(FacesContext context) {
		if (templateProviderFactory == null) {
			templateProviderFactory = createTemplateProviderFactory(context);
		}
		return templateProviderFactory;
	}

	public static TemplateProviderFactory getInstance() {
		final FacesContext facesContext = FacesContext.getCurrentInstance();
		if (facesContext == null) {
			throw new IllegalStateException("No Faces context found.");
		} else {
			return getInstance(facesContext);
		}
	}

	private static TemplateProviderFactory createTemplateProviderFactory(
			FacesContext context) {
		final String templateProviderFactoryClassName = context
				.getExternalContext()
				.getInitParameter(
						Constants.TEMPLATE_PROVIDER_FACTORY_CLASS_INIT_PARAMETER_NAME);

		if (templateProviderFactoryClassName == null) {
			return new DefaultTemplateProviderFactory();
		} else {
			try {
				final Class templateProviderFactoryClass = Class
						.forName(templateProviderFactoryClassName);
				final TemplateProviderFactory templateProviderFactory = (TemplateProviderFactory) templateProviderFactoryClass
						.newInstance();
				return templateProviderFactory;

			} catch (Exception ex) {
				logger
						.error(
								"Template provider factory ["
										+ templateProviderFactoryClassName
										+ "] could not be created. Creating the default template factory ["
										+ DefaultTemplateProviderFactory.class
												.getName() + "]", ex);
				return new DefaultTemplateProviderFactory();
			}
		}
	}
}
